/*
 * Date (yymmdd):               090320
 * Version:                     0.2
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */


#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pmodelchecker/pmodelchecker.h>

using namespace std;
using namespace ECSB;
using namespace PMODELCHECKER;


int main(int argc, char **argv)
{

  //Checking if the call is correct
  if(argc != 2)
    {
      cerr << "pmodelchecker xml_parameter_file" << std::endl;
      exit(EXIT_FAILURE);
    }

  //XML parameter file
  char* parameter_file = argv[1];

  //PRISM parameters
  MCParameters* MCParams = new MCParameters();

  //Read the parameter file
  MCParams->read(parameter_file);

  if(MCParams->model_checker == "PRISM")
  {
    //std::cout << "Task: " << MCParams->P_system_specification << endl;
    //Independently of the task to carry we translate our P system specifications
    //into a PRISM model.
    ECSB::CAPSystem Pmodel(MCParams->P_system_specification);
    PrismModel prismModel(Pmodel);

    //prismModel.show();
    prismModel.write(MCParams->PRISM_model);

    //XML file containing the variables of the model for communication with
    //property builder tool
    prismModel.writeParameters(MCParams->parameters_file);

    //PRISM command to perform the specific task
    std::string command;

    if (MCParams->task == "Build")
    {
        //Build the model by system calling prism PRISM_model parameters export_options
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
        command += MCParams->formula_parameters;
        command += " -exportstates ";
        command += MCParams->states_file;
        command += " -exporttrans ";
        command += MCParams->transitions_file;

        system(command.c_str());
    }
    else if (MCParams->task=="Verify")
    {
        //Verify properties specified in a file by system calling
        //prism model_file property_file parameters
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " ";
        command += MCParams->temporal_formulas;
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
        command += MCParams->formula_parameters;
        command += " -exportstates ";
        command += MCParams->states_file;
        command += " -exporttrans ";
        command += MCParams->transitions_file;
        command += " -exportresults ";
        command += MCParams->results_file;

        system(command.c_str());
    }
    else if (MCParams->task == "Approximate")
    {
        //Approximate properties specified in a file by system calling
        //prism model_file property_file -sim -simsamples
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " ";
        command += MCParams->temporal_formulas;
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
        command += MCParams->formula_parameters;
        command += " -sim ";
        command += " -simsamples ";
        command += somethingToString<unsigned long>(MCParams->number_samples);
        command += " -simconf ";
        command += somethingToString<float>(MCParams->confidence);
        command += " -simpathlen 10000000";
        command += " -exportresults ";
        command += MCParams->results_file;

        system(command.c_str());
    }

  }
  else if (MCParams->model_checker == "MC2")
  {
      if(MCParams->simulations_generated == "true")
      {
          generateSimulationsFileMC2(MCParams->simulations_file_hdf5,MCParams->number_samples,MCParams->simulations_file_MC2);
          //checkProperties(MCParams->simulations_file_MC2,MCParams->temporal_formulas,MCParams->results_file);          
      }
      else if(MCParams->simulations_generated == "false")
      {
          generateSimulationsFromModel(MCParams->mcss_params_file,MCParams->simulations_file_hdf5,MCParams->number_samples,MCParams->simulations_file_MC2);
          //checkProperties(MCParams->simulations_file_MC2,MCParams->temporal_formulas,MCParams->results_file);
      }
      else
      {
          cerr << "Unknown value for simulations_generated (true|false): " << MCParams->simulations_generated << std::endl;
          exit(EXIT_FAILURE);
      }

      checkProperties(MCParams->simulations_file_MC2,MCParams->temporal_formulas,MCParams->results_file);          



  }

  delete MCParams;

  return 0;
}

std::string obtainHead(std::string simulationsHDF5)
{
    //The command that outputs the head with [ to the file headFile.txt
    std::string command = "mcss-postprocess -l -t 1 " + simulationsHDF5 + " | head -n 1 > headFile.txt";

    //Execute the command invoking the system
    system(command.c_str());

    //Generate the head from the one in the headFile.txt
    std::string head;
    std::string fileName = "headFile.txt";
    ifstream headFile(fileName.c_str());

    if (headFile.is_open())
    {
        std::string line;
        while (! headFile.eof() )
        {
            getline (headFile,line);
            //cout << line << endl;
            head += line;
        }
        headFile.close();
    }
    else
    {
        cerr << "Unable to open file " << fileName << std::endl;
        exit(EXIT_FAILURE);
    }

    //Replacing [ with (
    size_t pos = head.find("[");

    while(pos != head.npos)
    {
        head.replace(pos,1,"(");
        pos = head.find("[",pos++);
    }

    //Replacing ] with )
    pos = head.find("]");

    while(pos != head.npos)
    {
        head.replace(pos,1,")");
        pos = head.find("]",pos++);
    }
 
    return head;
}


void generateSimulationsFileMC2(std::string simulationsHDF5,unsigned long runs,std::string simulationsMC2)
{

	 std::cout << "Generating simulations in MC2 format" << std::endl;

    //Generating the file in MC2 format containing the simulations
    std::string headInformation = obtainHead(simulationsHDF5);
    ofstream MC2file;

    //First simulation
    MC2file.open(simulationsMC2.c_str(),ios::out);
    MC2file << headInformation << std::endl;
    MC2file.close();

    //Generate command
    std::string command = "mcss-postprocess -p -l -t 1 " + simulationsHDF5 + " >> " + simulationsMC2;
    //std::cout << "Command to run: " << command << std::endl;
    system(command.c_str());


    for(int r = 1; r <= runs; ++r)
    {
        MC2file.open(simulationsMC2.c_str(),ios::app);
        MC2file << std::endl;
        MC2file << headInformation << std::endl;
        MC2file.close();

        //Generate command
        std::string command = "mcss-postprocess -p -l -t " + somethingToString<int>(r) + " " + simulationsHDF5 + " >> " + simulationsMC2;
        //std::cout << "Command to run: " << command << std::endl;
        system(command.c_str());
    }
}


void checkProperties(std::string simulationsFileMC2, std::string temporalFormulas, std::string resultFile)
{
    //Generate the expanded formulas
    //generateFormulas(temporalFormulas, "temporaryFile.txt");

    //Generate the command
    std::string command = "java -jar /MC2.jar stoch " + simulationsFileMC2 + " " + temporalFormulas + " 100 > " + resultFile;

    //Running MC2
    std::cout << "Running MC2" << std::endl;
    system(command.c_str());
}

void generateSimulationsFromModel(std::string mcssParamsFile, std::string HDF5file, unsigned long runs, std::string simulationsFileMC2)
{
    //Generate command to call mcss
    std::string command = "mcss " + mcssParamsFile + " data_file=" + HDF5file + " runs=" + somethingToString<unsigned long>(runs);

    //Call mcss using system
	 std::cout << "Running mcss" << std::endl;
    system(command.c_str());

    //Generate the file in MC2 format from the HDF5 file
    generateSimulationsFileMC2(HDF5file,runs,simulationsFileMC2);
}

